/*
 * Copyright (c) 2015, 2023, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package java.lang.invoke;

import jdk.internal.value.CheckedType;
import jdk.internal.value.NullRestrictedCheckedType;
import jdk.internal.value.ValueClass;
import jdk.internal.util.Preconditions;
import jdk.internal.vm.annotation.ForceInline;

import java.lang.invoke.VarHandle.VarHandleDesc;
import java.util.Objects;
import java.util.Optional;

import static java.lang.invoke.MethodHandleStatics.UNSAFE;

// -- This file was mechanically generated: Do not edit! -- //

final class VarHandleFlatValues {

    static sealed class FieldInstanceReadOnly extends VarHandle {
        final long fieldOffset;
        final Class<?> receiverType;
        final Class<?> fieldType;
        final CheckedType checkedFieldType;
        final int layout; // Unsafe.fieldLayout

        FieldInstanceReadOnly(Class<?> receiverType, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout) {
            this(receiverType, fieldOffset, fieldType, checkedFieldType, layout, FieldInstanceReadOnly.FORM, false);
        }

        protected FieldInstanceReadOnly(Class<?> receiverType, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout,
                                        VarForm form, boolean exact) {
            super(form, exact);
            this.fieldOffset = fieldOffset;
            this.receiverType = receiverType;
            this.fieldType = fieldType;
            this.checkedFieldType = checkedFieldType;
            this.layout = layout;
        }

        @Override
        public FieldInstanceReadOnly withInvokeExactBehavior() {
            return hasInvokeExactBehavior()
                ? this
                : new FieldInstanceReadOnly(receiverType, fieldOffset, fieldType, checkedFieldType, layout, vform, true);
        }

        @Override
        public FieldInstanceReadOnly withInvokeBehavior() {
            return !hasInvokeExactBehavior()
                ? this
                : new FieldInstanceReadOnly(receiverType, fieldOffset, fieldType, checkedFieldType, layout, vform, false);
        }

        @Override
        final MethodType accessModeTypeUncached(AccessType at) {
            return at.accessModeType(receiverType, fieldType);
        }

        @Override
        public Optional<VarHandleDesc> describeConstable() {
            var receiverTypeRef = receiverType.describeConstable();
            var fieldTypeRef = fieldType.describeConstable();
            if (!receiverTypeRef.isPresent() || !fieldTypeRef.isPresent())
                return Optional.empty();

            // Reflect on this VarHandle to extract the field name
            String name = VarHandles.getFieldFromReceiverAndOffset(
                receiverType, fieldOffset, fieldType).getName();
            return Optional.of(VarHandleDesc.ofField(receiverTypeRef.get(), name, fieldTypeRef.get()));
        }

        @ForceInline
        static Object get(VarHandle ob, Object holder) {
            FieldInstanceReadOnly handle = (FieldInstanceReadOnly)ob;
            Object value = UNSAFE.getFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getVolatile(VarHandle ob, Object holder) {
            FieldInstanceReadOnly handle = (FieldInstanceReadOnly)ob;
            Object value = UNSAFE.getFlatValueVolatile(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getOpaque(VarHandle ob, Object holder) {
            FieldInstanceReadOnly handle = (FieldInstanceReadOnly)ob;
            Object value = UNSAFE.getFlatValueOpaque(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getAcquire(VarHandle ob, Object holder) {
            FieldInstanceReadOnly handle = (FieldInstanceReadOnly)ob;
            Object value = UNSAFE.getFlatValueAcquire(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        static final VarForm FORM = new VarForm(FieldInstanceReadOnly.class, Object.class, Object.class);
    }

    static final class FieldInstanceReadWrite extends FieldInstanceReadOnly {
        FieldInstanceReadWrite(Class<?> receiverType, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout) {
            this(receiverType, fieldOffset, fieldType, checkedFieldType, layout, false);
        }

        private FieldInstanceReadWrite(Class<?> receiverType, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout,
                                       boolean exact) {
            super(receiverType, fieldOffset, fieldType, checkedFieldType, layout, FieldInstanceReadWrite.FORM, exact);
        }

        @Override
        public FieldInstanceReadWrite withInvokeExactBehavior() {
            return hasInvokeExactBehavior()
                ? this
                : new FieldInstanceReadWrite(receiverType, fieldOffset, fieldType, checkedFieldType, layout, true);
        }

        @Override
        public FieldInstanceReadWrite withInvokeBehavior() {
            return !hasInvokeExactBehavior()
                ? this
                : new FieldInstanceReadWrite(receiverType, fieldOffset, fieldType, checkedFieldType, layout, false);
        }

        @ForceInline
        static Object checkCast(FieldInstanceReadWrite handle, Object value) {
            return handle.checkedFieldType.cast(value);
        }

        @ForceInline
        static void set(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            UNSAFE.putFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                             handle.fieldOffset, handle.layout, handle.fieldType,
                             checkCast(handle, value));
        }

        @ForceInline
        static void setVolatile(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            UNSAFE.putFlatValueVolatile(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                     handle.fieldOffset, handle.layout, handle.fieldType,
                                     checkCast(handle, value));
        }

        @ForceInline
        static void setOpaque(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            UNSAFE.putFlatValueOpaque(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                   handle.fieldOffset, handle.layout, handle.fieldType,
                                   checkCast(handle, value));
        }

        @ForceInline
        static void setRelease(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            UNSAFE.putFlatValueRelease(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                    handle.fieldOffset, handle.layout, handle.fieldType,
                                    checkCast(handle, value));
        }

        @ForceInline
        static boolean compareAndSet(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.compareAndSetFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object compareAndExchange(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.compareAndExchangeFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object compareAndExchangeAcquire(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.compareAndExchangeFlatValueAcquire(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object compareAndExchangeRelease(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.compareAndExchangeFlatValueRelease(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetPlain(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.weakCompareAndSetFlatValuePlain(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSet(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.weakCompareAndSetFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetAcquire(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.weakCompareAndSetFlatValueAcquire(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetRelease(VarHandle ob, Object holder, Object expected, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.weakCompareAndSetFlatValueRelease(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSet(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.getAndSetFlatValue(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSetAcquire(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.getAndSetFlatValueAcquire(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSetRelease(VarHandle ob, Object holder, Object value) {
            FieldInstanceReadWrite handle = (FieldInstanceReadWrite)ob;
            return UNSAFE.getAndSetFlatValueRelease(Objects.requireNonNull(handle.receiverType.cast(holder)),
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        static final VarForm FORM = new VarForm(FieldInstanceReadWrite.class, Object.class, Object.class);
    }


    static sealed class FieldStaticReadOnly extends VarHandle {
        final Class<?> declaringClass;
        final Object base;
        final long fieldOffset;
        final Class<?> fieldType;
        final CheckedType checkedFieldType;
        final int layout;

        FieldStaticReadOnly(Class<?> declaringClass, Object base, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout) {
            this(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, FieldStaticReadOnly.FORM, false);
        }

        protected FieldStaticReadOnly(Class<?> declaringClass, Object base, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout,
                                      VarForm form, boolean exact) {
            super(form, exact);
            this.declaringClass = declaringClass;
            this.base = base;
            this.fieldOffset = fieldOffset;
            this.fieldType = fieldType;
            this.checkedFieldType = checkedFieldType;
            this.layout = layout;
        }

        @Override
        public FieldStaticReadOnly withInvokeExactBehavior() {
            return hasInvokeExactBehavior()
                ? this
                : new FieldStaticReadOnly(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, vform, true);
        }

        @Override
        public FieldStaticReadOnly withInvokeBehavior() {
            return !hasInvokeExactBehavior()
                ? this
                : new FieldStaticReadOnly(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, vform, false);
        }

        @Override
        public Optional<VarHandleDesc> describeConstable() {
            var fieldTypeRef = fieldType.describeConstable();
            if (!fieldTypeRef.isPresent())
                return Optional.empty();

            // Reflect on this VarHandle to extract the field name
            var staticField = VarHandles.getStaticFieldFromBaseAndOffset(
                declaringClass, fieldOffset, fieldType);
            var declaringTypeRef = declaringClass.describeConstable();
            if (!declaringTypeRef.isPresent())
                return Optional.empty();
            return Optional.of(VarHandleDesc.ofStaticField(declaringTypeRef.get(), staticField.getName(), fieldTypeRef.get()));
        }

        @Override
        final MethodType accessModeTypeUncached(AccessType at) {
            return at.accessModeType(null, fieldType);
        }

        @ForceInline
        static Object get(VarHandle ob) {
            FieldStaticReadOnly handle = (FieldStaticReadOnly) ob.target();
            Object value = UNSAFE.getFlatValue(handle.base,
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getVolatile(VarHandle ob) {
            FieldStaticReadOnly handle = (FieldStaticReadOnly) ob.target();
            Object value = UNSAFE.getFlatValueVolatile(handle.base,
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getOpaque(VarHandle ob) {
            FieldStaticReadOnly handle = (FieldStaticReadOnly) ob.target();
            Object value = UNSAFE.getFlatValueOpaque(handle.base,
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        @ForceInline
        static Object getAcquire(VarHandle ob) {
            FieldStaticReadOnly handle = (FieldStaticReadOnly) ob.target();
            Object value = UNSAFE.getFlatValueAcquire(handle.base,
                                 handle.fieldOffset, handle.layout, handle.fieldType);
            return value;
        }

        static final VarForm FORM = new VarForm(FieldStaticReadOnly.class, null, Object.class);
    }

    static final class FieldStaticReadWrite extends FieldStaticReadOnly {

        FieldStaticReadWrite(Class<?> declaringClass, Object base, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout) {
            this(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, false);
        }

        private FieldStaticReadWrite(Class<?> declaringClass, Object base, long fieldOffset, Class<?> fieldType, CheckedType checkedFieldType, int layout,
                                     boolean exact) {
            super(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, FieldStaticReadWrite.FORM, exact);
        }

        @Override
        public FieldStaticReadWrite withInvokeExactBehavior() {
            return hasInvokeExactBehavior()
                ? this
                : new FieldStaticReadWrite(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, true);
        }

        @Override
        public FieldStaticReadWrite withInvokeBehavior() {
            return !hasInvokeExactBehavior()
                ? this
                : new FieldStaticReadWrite(declaringClass, base, fieldOffset, fieldType, checkedFieldType, layout, false);
        }

        @ForceInline
        static Object checkCast(FieldStaticReadWrite handle, Object value) {
            return handle.checkedFieldType.cast(value);
        }

        @ForceInline
        static void set(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            UNSAFE.putFlatValue(handle.base,
                             handle.fieldOffset, handle.layout, handle.fieldType,
                             checkCast(handle, value));
        }

        @ForceInline
        static void setVolatile(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            UNSAFE.putFlatValueVolatile(handle.base,
                                     handle.fieldOffset, handle.layout, handle.fieldType,
                                     checkCast(handle, value));
        }

        @ForceInline
        static void setOpaque(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            UNSAFE.putFlatValueOpaque(handle.base,
                                   handle.fieldOffset, handle.layout, handle.fieldType,
                                   checkCast(handle, value));
        }

        @ForceInline
        static void setRelease(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            UNSAFE.putFlatValueRelease(handle.base,
                                    handle.fieldOffset, handle.layout, handle.fieldType,
                                    checkCast(handle, value));
        }

        @ForceInline
        static boolean compareAndSet(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.compareAndSetFlatValue(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }


        @ForceInline
        static Object compareAndExchange(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.compareAndExchangeFlatValue(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object compareAndExchangeAcquire(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.compareAndExchangeFlatValueAcquire(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object compareAndExchangeRelease(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.compareAndExchangeFlatValueRelease(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetPlain(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.weakCompareAndSetFlatValuePlain(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSet(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.weakCompareAndSetFlatValue(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetAcquire(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.weakCompareAndSetFlatValueAcquire(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static boolean weakCompareAndSetRelease(VarHandle ob, Object expected, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.weakCompareAndSetFlatValueRelease(handle.base,
                                               handle.fieldOffset, handle.layout, handle.fieldType,
                                               checkCast(handle, expected),
                                               checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSet(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.getAndSetFlatValue(handle.base,
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSetAcquire(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.getAndSetFlatValueAcquire(handle.base,
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        @ForceInline
        static Object getAndSetRelease(VarHandle ob, Object value) {
            FieldStaticReadWrite handle = (FieldStaticReadWrite) ob.target();
            return UNSAFE.getAndSetFlatValueRelease(handle.base,
                                          handle.fieldOffset, handle.layout, handle.fieldType,
                                          checkCast(handle, value));
        }

        static final VarForm FORM = new VarForm(FieldStaticReadWrite.class, null, Object.class);
    }

    static final class Array extends VarHandle {
        final int abase;
        final int ashift;
        final Class<?> arrayType;
        final Class<?> componentType;
        final int layout;

        Array(int abase, int ashift, Class<?> arrayType, int layout) {
            this(abase, ashift, arrayType, layout, false);
        }

        private Array(int abase, int ashift, Class<?> arrayType, int layout, boolean exact) {
            super(Array.FORM, exact);
            this.abase = abase;
            this.ashift = ashift;
            this.arrayType = arrayType;
            this.componentType = arrayType.getComponentType();
            this.layout = layout;
        }

        @Override
        public Array withInvokeExactBehavior() {
            return hasInvokeExactBehavior()
                ? this
                : new Array(abase, ashift, arrayType, layout, true);
        }

        @Override
        public Array withInvokeBehavior() {
            return !hasInvokeExactBehavior()
                ? this
                : new Array(abase, ashift, arrayType, layout, false);
        }

        @Override
        public Optional<VarHandleDesc> describeConstable() {
            var arrayTypeRef = arrayType.describeConstable();
            if (!arrayTypeRef.isPresent())
                return Optional.empty();

            return Optional.of(VarHandleDesc.ofArray(arrayTypeRef.get()));
        }

        @Override
        final MethodType accessModeTypeUncached(AccessType at) {
            return at.accessModeType(arrayType, componentType, int.class);
        }

        @ForceInline
        static Object runtimeTypeCheck(Array handle, Object[] oarray, Object value) {
            if (handle.arrayType == oarray.getClass()) {
                // Fast path: static array type same as argument array type
                return ValueClass.componentCheckedType(oarray).cast(value);
            } else {
                // Slow path: check value against argument array component checked type
                return reflectiveTypeCheck(oarray, value);
            }
        }

        @ForceInline
        static Object reflectiveTypeCheck(Object[] oarray, Object value) {
            try {
                return ValueClass.componentCheckedType(oarray).cast(value);
            } catch (ClassCastException e) {
                throw new ArrayStoreException();
            }
        }

        @ForceInline
        static Object get(VarHandle ob, Object oarray, int index) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return array[index];
        }

        @ForceInline
        static void set(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            array[index] = runtimeTypeCheck(handle, array, value);
        }

        @ForceInline
        static Object getVolatile(VarHandle ob, Object oarray, int index) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getFlatValueVolatile(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType);
        }

        @ForceInline
        static void setVolatile(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            UNSAFE.putFlatValueVolatile(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object getOpaque(VarHandle ob, Object oarray, int index) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getFlatValueOpaque(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType);
        }

        @ForceInline
        static void setOpaque(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            UNSAFE.putFlatValueOpaque(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object getAcquire(VarHandle ob, Object oarray, int index) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getFlatValueAcquire(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType);
        }

        @ForceInline
        static void setRelease(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            UNSAFE.putFlatValueRelease(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static boolean compareAndSet(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.compareAndSetFlatValue(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object compareAndExchange(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.compareAndExchangeFlatValue(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object compareAndExchangeAcquire(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.compareAndExchangeFlatValueAcquire(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object compareAndExchangeRelease(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.compareAndExchangeFlatValueRelease(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static boolean weakCompareAndSetPlain(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.weakCompareAndSetFlatValuePlain(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static boolean weakCompareAndSet(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.weakCompareAndSetFlatValue(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static boolean weakCompareAndSetAcquire(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.weakCompareAndSetFlatValueAcquire(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static boolean weakCompareAndSetRelease(VarHandle ob, Object oarray, int index, Object expected, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.weakCompareAndSetFlatValueRelease(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    handle.componentType.cast(expected),
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object getAndSet(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getAndSetFlatValue(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object getAndSetAcquire(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getAndSetFlatValueAcquire(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }

        @ForceInline
        static Object getAndSetRelease(VarHandle ob, Object oarray, int index, Object value) {
            Array handle = (Array)ob;
            Object[] array = (Object[]) handle.arrayType.cast(oarray);
            return UNSAFE.getAndSetFlatValueRelease(array,
                    (((long) Preconditions.checkIndex(index, array.length, Preconditions.AIOOBE_FORMATTER)) << handle.ashift) + handle.abase, handle.layout, handle.componentType,
                    runtimeTypeCheck(handle, array, value));
        }
        static final VarForm FORM = new VarForm(Array.class, Object[].class, Object.class, int.class);
    }
    static final ClassValue<Array> flatArrayVarHandles = new ClassValue<>() {
        @Override protected Array computeValue(Class<?> arrayClass) {
            assert UNSAFE.isFlatArray(arrayClass);
            int aoffset = (int) UNSAFE.arrayBaseOffset(arrayClass);
            int ascale = UNSAFE.arrayIndexScale(arrayClass);
            int ashift = 31 - Integer.numberOfLeadingZeros(ascale);
            int layout = UNSAFE.arrayLayout(arrayClass);
            return new Array(aoffset, ashift, arrayClass, layout);
        }
    };
    static VarHandle flatArrayVarHandle(Class<?> arrayClass) {
        return flatArrayVarHandles.get(arrayClass);
    }
}
